/*
use this config to set the port/host that php is listening on and to set the root directory that php will use for resolving script paths
*/
exports.options = {
	port: "/tmp/php.sock",
	host: null,
	root: "/source/db/php-js/phpmyadmin",
	server: {
		host: "10.11.12.75",
		port: 8124
	}
};